from __future__ import absolute_import

from .interface import AbstractHmmParams, AbstractDagParams
from .implement import DefaultHmmParams, DefaultDagParams
from .priorityset import Item, PrioritySet
from .util import is_chinese, remove_tone, normlize_pinyin, simplify_pinyin, is_pinyin, all_pinyin


from .dag import dag
from .viterbi import viterbi


# coding: utf-8
from __future__ import (print_function, unicode_literals, absolute_import)

import os
import sys

try:
    reload(sys)
    sys.setdefaultencoding('utf-8')
except:
    pass


PY2 = sys.version_info[0] == 2
if not PY2:
    # Python 3.x and up
    xrange       = range

    def as_text(v):  ## 生成unicode字符串
        if v is None:
            return None
        elif isinstance(v, bytes):
            return v.decode('utf-8', errors='ignore')
        elif isinstance(v, str):
            return v
        else:
            raise ValueError('Unknown type %r' % type(v))

    def is_text(v):
        return isinstance(v, str)

else:
    # Python 2.x
    xrange       = xrange

    def as_text(v):
        if v is None:
            return None
        elif isinstance(v, unicode):
            return v
        elif isinstance(v, str):
            return v.decode('utf-8', errors='ignore')
        else:
            raise ValueError('Invalid type %r' % type(v))

    def is_text(v):
        return isinstance(v, unicode)


def is_chinese(v):
    if is_text(v):
        if len(v) == 0:
            return False
        return all(u'\u4e00' <= c <= u'\u9fff' or c == u'〇' for c in v)
    else:
        raise ValueError('Invalid type %r' % type(v))

def current_dir():
    return os.path.dirname(os.path.realpath(__file__))


__removetone_dict = {
    'ā': 'a',
    'á': 'a',
    'ǎ': 'a',
    'à': 'a',
    'ē': 'e',
    'é': 'e',
    'ě': 'e',
    'è': 'e',
    'ī': 'i',
    'í': 'i',
    'ǐ': 'i',
    'ì': 'i',
    'ō': 'o',
    'ó': 'o',
    'ǒ': 'o',
    'ò': 'o',
    'ū': 'u',
    'ú': 'u',
    'ǔ': 'u',
    'ù': 'u',
    'ü': 'v',
    'ǖ': 'v',
    'ǘ': 'v',
    'ǚ': 'v',
    'ǜ': 'v',
    'ń': 'n',
    'ň': 'n',
    '': 'm',
}

## 由于从 __future__ 导入了 unicode_literals ， 所以字符串默认是unicode的
# for k in __removetone_dict.keys():
#     v = __removetone_dict[k]
#     __removetone_dict.pop(k, None)
#     __removetone_dict[as_text(k)] = as_text(v)

def remove_tone(one_py):
    """ 删除拼音中的音调
    lǔ -> lu
    """
    one_py = as_text(one_py)
    r = as_text('')
    for c in one_py:
        if c in __removetone_dict:
            r += __removetone_dict[c]
        else:
            r += c
    return r


def normlize_pinyin(one_py):
    """ 规范化
    ue -> ve
    """
    if 'ue' in one_py:
        return one_py.replace('ue', 've')
    if 'ng' == one_py:   # 嗯
        return 'en'
    return one_py

def simplify_pinyin(one_py):
    return normlize_pinyin( remove_tone(one_py.lower()) )

# 拼音
__pinyin = set(['gu','qiao','qian','qve','ge','gang','ga','lian','liao','rou','zong',\
    'tu','seng','yve','ti','te','jve','ta','nong','zhang','fan','ma','gua','die','gui',\
    'guo','gun','sang','diu','zi','ze','za','chen','zu','ba','dian','diao','nei','suo',\
    'sun','zhao','sui','kuo','kun','kui','cao','zuan','kua','den','lei','neng','men',\
    'mei','tiao','geng','chang','cha','che','fen','chi','fei','chu','shui','me','tuan',\
    'mo','mi','mu','dei','cai','zhan','zhai','can','ning','wang','pie','beng','zhuang',\
    'tan','tao','tai','song','ping','hou','cuan','lan','lao','fu','fa','jiong','mai',\
    'xiang','mao','man','a','jiang','zun','bing','su','si','sa','se','ding','xuan',\
    'zei','zen','kong','pang','jie','jia','jin','lo','lai','li','peng','jiu','yi','yo',\
    'ya','cen','dan','dao','ye','dai','zhen','bang','nou','yu','weng','en','ei','kang',\
    'dia','er','ru','keng','re','ren','gou','ri','tian','qi','shua','shun','shuo','qun',\
    'yun','xun','fiao','zan','zao','rang','xi','yong','zai','guan','guai','dong','kuai',\
    'ying','kuan','xu','xia','xie','yin','rong','xin','tou','nian','niao','xiu','fo',\
    'kou','niang','hua','hun','huo','hui','shuan','quan','shuai','chong','bei','ben',\
    'kuang','dang','sai','ang','sao','san','reng','ran','rao','ming','null','lie','lia',\
    'min','pa','lin','mian','mie','liu','zou','miu','nen','kai','kao','kan','ka','ke',\
    'yang','ku','deng','dou','shou','chuang','nang','feng','meng','cheng','di','de','da',\
    'bao','gei','du','gen','qu','shu','sha','she','ban','shi','bai','nun','nuo','sen','lve',\
    'kei','fang','teng','xve','lun','luo','ken','wa','wo','ju','tui','wu','le','ji','huang',\
    'tuo','cou','la','mang','ci','tun','tong','ca','pou','ce','gong','cu','lv','dun','pu',\
    'ting','qie','yao','lu','pi','po','suan','chua','chun','chan','chui','gao','gan','zeng',\
    'gai','xiong','tang','pian','piao','cang','heng','xian','xiao','bian','biao','zhua','duan',\
    'cong','zhui','zhuo','zhun','hong','shuang','juan','zhei','pai','shai','shan','shao','pan',\
    'pao','nin','hang','nie','zhuai','zhuan','yuan','niu','na','miao','guang','ne','hai','han',\
    'hao','wei','wen','ruan','cuo','cun','cui','bin','bie','mou','nve','shen','shei','fou','xing',\
    'qiang','nuan','pen','pei','rui','run','ruo','sheng','dui','bo','bi','bu','chuan','qing',\
    'chuai','duo','o','chou','ou','zui','luan','zuo','jian','jiao','sou','wan','jing','qiong',\
    'wai','long','yan','liang','lou','huan','hen','hei','huai','shang','jun','hu','ling','ha','he',\
    'zhu','ceng','zha','zhe','zhi','qin','pin','ai','chai','qia','chao','ao','an','qiu','ni','zhong',\
    'zang','nai','nan','nao','chuo','tie','you','nu','nv','zheng','leng','zhou','lang','e',])

# 声母
__shengmu = set(['b','p','m','f','d','t','n','l','g','k','h','j','q','x','zh','ch','sh','r','z','c','s',])
# 韵母
__single_yunmu = set(['a', 'o', 'e', 'i', 'u', 'v'])
__complex_yunmu = set(['ai','ei','ui','ao','ou','iu','ie','ve','er','an','en','in','un','ang','eng','ing','ong',])

def is_pinyin(v):
    return v in __pinyin

def all_pinyin():
    for _ in __pinyin:
        yield _

def is_shengmu(v):
    return v in __shengmu

def is_single_yunmu(v):
    return v in __single_yunmu

def is_complex_yunmu(v):
    return v in __complex_yunmu

def is_yunmu(v):
    return is_single_yunmu(v) or is_complex_yunmu(v)

def get_shengmu(one_py):
    if len(one_py) == 0:
        return None
    elif len(one_py) == 1:
        if is_shengmu(one_py):
            return one_py
        else:
            return None
    else:
        if is_shengmu(one_py[:2]):
            return one_py[:2]
        elif is_shengmu(one_py[:1]):
            return one_py[:1]
        else:
            return None


# coding: utf-8
from __future__ import (print_function, unicode_literals, absolute_import)

from .interface import AbstractHmmParams, AbstractDagParams
from .util import as_text
import os
import json


DATA    = 'data'
DEFAULT = 'default'

class DefaultHmmParams(AbstractHmmParams):

    def __init__(self,):
        current_dir = self.pwd()
        self.py2hz_dict      = self.readjson(os.path.join(current_dir, 'data', 'hmm_py2hz.json'))
        self.start_dict      = self.readjson(os.path.join(current_dir, 'data', 'hmm_start.json'))
        self.emission_dict   = self.readjson(os.path.join(current_dir, 'data', 'hmm_emission.json'))
        self.transition_dict = self.readjson(os.path.join(current_dir, 'data', 'hmm_transition.json'))

    def readjson(self, filename):
        with open(filename) as outfile:
            return json.load(outfile)

    def pwd(self):
        return os.path.dirname(os.path.abspath(__file__))

    def start(self, state):
        ''' get start prob of state(hanzi) '''
        state = as_text(state)

        data = self.start_dict[DATA]
        default = self.start_dict[DEFAULT]

        if state in data:
            prob = data[state]
        else:
            prob = default
        return float(prob)


    def emission(self, state, observation):
        ''' state (hanzi) -> observation (pinyin) '''
        pinyin = as_text(observation)
        hanzi = as_text(state)

        data = self.emission_dict[DATA]
        default = self.emission_dict[DEFAULT] 

        if hanzi not in data:
            return float( default )
        
        prob_dict = data[hanzi]

        if pinyin not in prob_dict:
            return float( default )
        else:
            return float( prob_dict[pinyin] )

    def transition(self, from_state, to_state):
        ''' state -> state '''
        from_state = as_text(from_state)
        to_state = as_text(to_state)
        prob = 0.0

        data = self.transition_dict[DATA]
        default = self.transition_dict[DEFAULT]

        if from_state not in data:
            return float( default )
        
        prob_dict = data[from_state]

        if to_state in prob_dict:
            return float( prob_dict[to_state] )
        
        if DEFAULT in prob_dict:
            return float( prob_dict[DEFAULT] )

        return float( default )

    def get_states(self, observation):
        ''' get states which produce the given obs '''
        return [hanzi for hanzi in self.py2hz_dict[observation]]



class DefaultDagParams(AbstractDagParams):

    def __init__(self,):
        current_dir = self.pwd()
        self.char_dict      = self.readjson(os.path.join(current_dir, 'data', 'dag_char.json'))
        self.phrase_dict    = self.readjson(os.path.join(current_dir, 'data', 'dag_phrase.json'))

    def readjson(self, filename):
        with open(filename) as outfile:
            return json.load(outfile)

    def pwd(self,):
        return os.path.dirname(os.path.abspath(__file__))

    def get_phrase(self, pinyin_list, num=6):
        ''' pinyin_list是拼音组成的list，例如['yi', 'ge'] '''
        if len(pinyin_list) == 0:
            return []
        if len(pinyin_list) == 1:
            data = self.char_dict
        else:
            data = self.phrase_dict

        pinyin = ','.join(pinyin_list)

        if pinyin not in data:
            return []

        return data[pinyin][:num]
        




# coding: utf-8

import heapq

class Item(object):

    def __init__(self, score, path):
        self.__score = score
        self.__path  = path

    @property
    def score(self):
        return self.__score

    @property
    def path(self):
        return self.__path

    def __lt__(self, other):
        return self.__score < other.score

    def __le__(self, other):
        return self.__score <= other.score

    def __eq__(self, other):
        return self.__score == other.score

    def __ne__(self, other):
        return self.__score != other.score

    def __gt__(self, other):
        return self.__score > other.score

    def __ge__(self, other):
        return self.__score >= other.score

    def __str__(self):
        return '< score={0}, path={1} >'.format(self.__score, self.__path)

    def __repr__(self):
        return self.__str__()


class PrioritySet(object):

    def __init__(self, capacity):
        self.capacity = capacity
        self.data = []

    def put(self, score, path):
        assert(isinstance(path, list) == True)
        heapq.heappush(self.data, [score, Item(score, path)])
        while len(self.data) > self.capacity:
            heapq.heappop(self.data)

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        for item in self.data:
            yield item[1]

    def __str__(self):
        s = '[ \n'
        for item in self.data:
            s = s + '\t' + str(item[1]) + '\n'
        s += ']'
        return s

    def __repr__(self):
        return self.__str__()



# coding: utf-8
from __future__ import (print_function, unicode_literals, absolute_import)

from .interface import AbstractDagParams
from .priorityset import PrioritySet
from .util import xrange
import math

def dag(dag_params, pinyin_list, path_num=6, log=False):

    assert( isinstance(dag_params, AbstractDagParams) )

    pinyin_num = len(pinyin_list)
    if pinyin_num == 0:
        return []

    D = [PrioritySet(path_num) for _ in xrange(pinyin_num)]

    ## idx is 1
    for from_idx in xrange(0, 1):
        for to_idx in xrange(from_idx, pinyin_num):
            kvs = dag_params.get_phrase(pinyin_list[from_idx:to_idx+1], num=path_num)
            for item in kvs:
                word = [item[0]]
                if log:
                    score = math.log(item[1])
                else:
                    score = item[1]
                D[to_idx].put(score, word)

    for from_idx in xrange(1, pinyin_num):
        prev_paths = D[from_idx-1]
        for to_idx in xrange(from_idx, pinyin_num):
            kvs = dag_params.get_phrase(pinyin_list[from_idx:to_idx+1], num=path_num)
            for prev_item in prev_paths:
                for item in kvs:
                    word = prev_item.path + [ item[0] ]
                    if log:
                        score = prev_item.score + math.log(item[1])
                    else:
                        score = prev_item.score * item[1]
                    D[to_idx].put(score, word)

    result = [ item for item in D[-1] ]

    return sorted(result, key=lambda item: item.score, reverse=True)











# coding: utf-8
from __future__ import (print_function, unicode_literals, absolute_import)

from .interface import AbstractHmmParams
from .priorityset import PrioritySet
import math

def viterbi(hmm_params, observations, path_num=6, log=False, min_prob=3.14e-200):

    assert( isinstance(hmm_params, AbstractHmmParams) )

    V = [{}]
    t = 0
    cur_obs = observations[t]
    
    # Initialize base cases (t == 0)
    prev_states = cur_states = hmm_params.get_states(cur_obs)  # wordset
    for state in cur_states:
        if log:
            __score   = math.log(max(hmm_params.start(state), min_prob)) + \
                math.log(max(hmm_params.emission(state, cur_obs), min_prob))
        else:
            __score   = max(hmm_params.start(state), min_prob) * \
                max(hmm_params.emission(state, cur_obs), min_prob)
        __path    = [state]
        V[0].setdefault(state, PrioritySet(path_num))
        V[0][state].put(__score, __path)

    
    # Run Viterbi for t > 0
    for t in range(1, len(observations)):
        cur_obs = observations[t]

        if len(V) == 2:
            V = [V[-1]]

        V.append({})

        prev_states = cur_states
        cur_states = hmm_params.get_states(cur_obs)

        for y in cur_states:
            V[1].setdefault( y, PrioritySet(path_num) )
            max_item = None
            for y0 in prev_states:  # from y0(t-1) to y(t)
                for item in V[0][y0]:
                    if log:
                        _s = item.score + \
                            math.log(max(hmm_params.transition(y0, y), min_prob)) + \
                            math.log(max(hmm_params.emission(y, cur_obs), min_prob))
                    else:
                        _s = item.score * \
                            max(hmm_params.transition(y0, y), min_prob) * \
                            max(hmm_params.emission(y, cur_obs), min_prob)

                    _p = item.path + [y]
                    V[1][y].put(_s, _p)

    result = PrioritySet(path_num)
    for last_state in V[-1]:
        for item in V[-1][last_state]:
            result.put(item.score, item.path)
    result = [item for item in result]

    return sorted(result, key=lambda item: item.score, reverse=True)


# coding: utf-8

class AbstractHmmParams(object):
    
    def start(self, state):
        ''' get start prob of state(hanzi) '''
        pass

    def emission(self, state, observation):
        ''' state (hanzi) -> observation (pinyin) '''
        pass

    def transition(self, from_state, to_state):
        ''' state -> state '''
        pass

    def get_states(self, observation):
        ''' get states which produce the given obs '''
        pass

class AbstractDagParams(object):

    def get_phrase(self, pinyin_list, num):
        pass